// *** 3D Object Viewing Screen Saver Program ***
// *** Written by Steven De Toni Febyary 1996 ***
// **********************************************

#include <windows.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <commdlg.h>
#include <dos.h>        // gettime(), time structure, disable(), enable()

// ***** Include 3D object information *****
#include "ball3d.h"     // Ball3D object definition.
#include "config.h"     // Configuration file define constants
#include "ctl3d.h"      // Fancey look to dialog boxes in Win3.1.


// Structure used to determine ZDistance of objects. This is a basic
// painters algorithm implementation when drawing objects. That is to
// say if two object colide with each other, one will over write the
// other.
struct ObjDistance
{
    int  objIndx;
    long objZdistance;
};

// Program varibales !
const  char*      pSzProgName = "3D Screen Saver"; // Program name.
const  int        StrSize     = 256;               // String size for string processing.
static HINSTANCE  HInst;                           // Handle to main window of application.

// *** Prototypes ***
long FAR PASCAL _export    WndProc              (HWND hWnd, WORD message, WORD wParam, LONG lParam);
BOOL FAR PASCAL _export    ConfigDialog         (HWND hDlg, WORD message, WORD wParam, LONG lParam);
BOOL FAR PASCAL _export    SetPasswordDialog    (HWND hDlg, WORD message, WORD wParam, LONG lParam);
BOOL FAR PASCAL _export    GetUserPDialog       (HWND hDlg, WORD message, WORD wParam, LONG lParam);
LRESULT FAR PASCAL _export KeyboardProc         (int code, WPARAM wParam, LPARAM lParam);

// Spinner and field entry limits in configuration dialog.
const int RotationMax    = 90,   RotationMin    = -90;
const int TranslationMax = 50,   TranslationMin  = 1;
const int ZDistMax       = 2000, ZDistMin        = 700;
const int ObjNoMax       = 10000,ObjNoMin        = 1;
const int FrameRate      = 1;

// Default INI settings.
const char* Section        = "Configuration";
const char* DefIniSettings = "1 1 1 5 5 5 800 1 pyramid.3d ~ 0";

// Timer ID constant.
const int TIMERID = 10;  // Timer ID within function.

// Time limit variables for entering password.
const  long PASSWORDCYCLIMIT = 3000; // Timit limit to enter password before dialod is removed.
static long CurrentCycle     = 0;
static HWND HPassDialog      = NULL; // Varible used to commincate with
                                     // Password dialog box, and close it
                                     // after a certain time. If variable is
                                     // not null, then password dialog box is
                                     // open.


// Current system screen size.
int MAXX, MAXY;

// 3D Object information used in configure each object before being
// displayed. These variables are stored in the INI file (win.ini)
static int   XRot,      YRot,   ZRot  ;     // X,Y,Z angles of object
static int   XTrans,    YTrans, ZTrans;     // X,Y,Z movement.
static int   ZDistance, ObjectNum;
static char  ObjFilename[StrSize];
static char  Password   [StrSize/5];
static int   UsePassword;


// ***************************************************************************************
// Program Entry, and exit.
// ***************************************************************************************
int PASCAL WinMain  (HANDLE hInstance, HANDLE hPrevInstance, LPSTR lpszCmdParam, int nCmdShow)
{
    HWND            hwnd;
    MSG             msg;
    WNDCLASS        wndclass;
    char            paramBuff[StrSize];

    HInst = hInstance;                                  // Assign program instance

    if (!hPrevInstance)
    {
        wndclass.style                = CS_BYTEALIGNWINDOW | CS_HREDRAW | CS_VREDRAW;
        wndclass.lpfnWndProc          = (WNDPROC) WndProc;
        wndclass.cbClsExtra           = 0;
        wndclass.hInstance            = HInst;
        wndclass.hIcon                = NULL;
        wndclass.hCursor              = LoadCursor (HInst, MAKEINTRESOURCE(BLANKCURSOR));
        wndclass.hbrBackground        = GetStockObject (BLACK_BRUSH);
        wndclass.lpszMenuName         = NULL;
        wndclass.lpszClassName        = pSzProgName;
        
        RegisterClass (&wndclass);      
    }
    else
        return 0;   // Don't more than one instance of the screen saver

    MAXX  = GetSystemMetrics(SM_CXSCREEN);
    MAXY  = GetSystemMetrics(SM_CYSCREEN);

    // Check parameter string for configuration command /c
    lstrcpy (paramBuff, lpszCmdParam);
    strupr (paramBuff);
    if (strstr (paramBuff, "/C") != NULL)
    {       
        // Setup the control 3D sub-classing DLL.
        Ctl3dRegister(HInst);
        Ctl3dAutoSubclass(HInst);

        // Test Configuration Dialog.
        static FARPROC fpfnConfigDialog;
        fpfnConfigDialog = MakeProcInstance ((FARPROC)ConfigDialog, HInst);
        DialogBox (HInst, MAKEINTRESOURCE(IDD_CONFIGDIALOG), NULL, (DLGPROC) fpfnConfigDialog);
        FreeProcInstance (fpfnConfigDialog);

        // Free up usage count of 3D DLL.
        Ctl3dUnregister(HInst);
        return 0;
    }

    // Create Screen Saver window.
    hwnd  = CreateWindowEx (WS_EX_TOPMOST,
                         pSzProgName,                   // Window class name
                         "",                            // Window caption
                         WS_POPUP,                      // Window style
	                     0,                             // inital X position
                         0,                             // inital y position 
	                     MAXX,                          // inital X size
                         MAXY,                          // intial y size
                         NULL,                          // parent window handle
                         NULL,                          // window menu handle
                         HInst,                         // program instance handle
                         NULL);                         // creation parameters


    // Go into screen saver mode. 
    ShowWindow (hwnd, nCmdShow);
    UpdateWindow (hwnd);

    // Setup Screen saver timer.
    SetTimer(hwnd, TIMERID, FrameRate, (TIMERPROC) WndProc);

    // Setup hook funcmtion to intercept key presses
    // (**** This doesn't capture cntl alt del keys !!! ****)
    static HHOOK    hKeyProc;
    static HOOKPROC fpKeyProc;
    fpKeyProc = MakeProcInstance ((FARPROC) KeyboardProc, HInst); 
    hKeyProc = SetWindowsHookEx(WH_KEYBOARD, fpKeyProc, HInst, GetCurrentTask());

    // Process screen saver messages...
    while (GetMessage (&msg, NULL, 0, 0))
    {
        
        TranslateMessage (&msg);
        DispatchMessage  (&msg);
    }

    // Free hook function resources.
    UnhookWindowsHookEx(hKeyProc);
    FreeProcInstance (fpKeyProc);

    return msg.wParam;
}

// To custom implement sort routine, change the type of the following
// variables: pData[], midRange, temp.
//
// Alter comparsion testing for the following statements:
// while (pData[lowInc] < midRange)
// while (pData[highDec] > midRange)
// Alter swap area.
void QuickSort (ObjDistance pData[], long lowRange, long highRange)
{
    long lowInc   = lowRange;
    long highDec  = highRange;
    long midRange = pData[(lowRange+highRange) >> 1].objZdistance;
    do
    {
        // test elements from lower to midrange
        while (pData[lowInc].objZdistance < midRange)
              lowInc++;

        // test elements from higher to midrange
        while (pData[highDec].objZdistance > midRange)
              highDec--;

        if (lowInc <= highDec)
        {
            // Swap values
            ObjDistance temp = pData[lowInc];
            pData[lowInc]    = pData[highDec];
            pData[highDec]   = temp;

            // move to test next items
            lowInc++;
            highDec--;
        }

    } while (highDec > lowInc);

    // Recursive calling to sort other parts of array
    if (lowRange < highDec)
       QuickSort (pData, lowRange, highDec);

    if (lowInc < highRange)
       QuickSort (pData, lowInc, highRange);
}


// Unused.
void WasteCPUTime (int seconds)
{
    struct time startTime;
    struct time endTime;

    gettime (&startTime);
    do
    {
        gettime (&endTime);
    }
    while (abs(endTime.ti_sec - startTime.ti_sec) < seconds);
}

// Hook function to process keys that user has pressed
LRESULT FAR PASCAL _export KeyboardProc (int code, WPARAM wParam, LPARAM lParam)
{
    /****   switch (wParam)
    {
        case VK_CONTROL:
        case VK_MENU:
             if (HPassDialog != NULL)
             {
                 MessageBeep (MB_ICONINFORMATION);
                 disable();
                 WasteCPUTime (5);
                 enable();
                 return 1;
             }
    }
    ****/
    return 0;
}

// ***************************************************************************
// *** Get User Password Dialog Box. Open up the dialog box to recieve the
// *** user's password.
// ***************************************************************************
BOOL FAR PASCAL _export GetUserPDialog (HWND hDlg, WORD message, WORD wParam, LONG lParam)
{
	switch (message)
    {
        case WM_INITDIALOG:
             HPassDialog  = hDlg;
             CurrentCycle = 0;
             SetActiveWindow (GetDlgItem(hDlg, IDC_GETPASSWORD));
             return TRUE;

        case WM_CHAR:
             MessageBeep (-1);
             break;

        case WM_KEYDOWN:
        case WM_SYSKEYDOWN:
             return TRUE;

        case WM_COMMAND:    // A user has selected a control command
            switch (LOWORD (wParam))
            {
                case IDC_CANCEL:
                    EndDialog (hDlg, FALSE);
                    HPassDialog = NULL;
                    return TRUE;
                    
                case IDC_OK:
                     char testPass[StrSize/5];

                     GetDlgItemText(hDlg, IDC_GETPASSWORD, testPass, StrSize/5);
                     if (strcmp (Password, testPass) != 0)
                     {
                        MessageBeep(MB_ICONINFORMATION);
                        SetWindowText (hDlg,"Password Incorrect. Retry Entry...");
                        SetActiveWindow (GetDlgItem(hDlg, IDC_GETPASSWORD));
                        break;
                     }

                     // If all checks pass ... exit!
                     EndDialog (hDlg, TRUE);
                     HPassDialog = NULL;
                     return TRUE;

				default:
                	break;
            }     
            break;

        case WM_CLOSE:
			EndDialog (hDlg, 0);
            HPassDialog = NULL;
            return TRUE;
    }
              
	return FALSE;
}


// ***************************************************************************
// *** Set Password Dialog Box.
// *** Get user's password from within the configuration dialog box.
// ***************************************************************************
BOOL FAR PASCAL _export SetPasswordDialog (HWND hDlg, WORD message, WORD wParam, LONG lParam)
{
	switch (message)
    {
        case WM_COMMAND:    // A user has selected a control command
            switch (LOWORD (wParam))
            {
                case IDC_SETPASSWORD:
                     char newPass[StrSize/5];
                     char testPass[StrSize/5];

                     GetDlgItemText(hDlg, IDC_NEWPASSWORD, newPass, StrSize/5);
                     if (strlen (newPass) <= 0)
                     {
                        MessageBox (hDlg, "Please Enter A Password First", "*** Entry Error ***",
                                    MB_OK | MB_ICONINFORMATION);
                        break;
                     }

                     GetDlgItemText(hDlg, IDC_RETYPEDPASSWORD, testPass, StrSize/5);
                     if (strcmp (newPass, testPass) != 0)
                     {
                        MessageBox (hDlg, "New And Retyped Password Differ", "*** Check Error ***",
                                    MB_OK | MB_ICONINFORMATION);
                        break;
                     }

                     // If all checks pass ... then use it !
                     UsePassword = 1;
                     strcpy (Password, testPass);
                     EndDialog (hDlg, 1);
                     break;

                case IDC_CANCELPASSWORD:
  					 EndDialog (hDlg, 0);
				     break;

				default:
                	break;
            }     
            break;

        case WM_CLOSE:
			EndDialog (hDlg, 0);
            break;
    }
              
	return FALSE;
}



// **********************************************************************************
// Updated process spinner controls (aka: scroll bars).
// 
// hDlg        = handle to dialog box
// cntlID      = ID number of text field
// direction   = 1, for up, -1 for down, 0 = dont increment/decrement
// heightLimit = max number (inclusive).
// lowlimit    = min number (inclusive).
//
// Returns integer value of field.
int UpdateField (HWND hDlg, int cntlID, int direction, int heightLimit, int lowLimit)
{
    char buffer[StrSize];
    int  number;

    // Get text
    GetDlgItemText(hDlg, cntlID, buffer, StrSize);

    // convert string.
    number = atoi (buffer);
    
    // Check field constraints.
    if ((direction >= 1) && (number < heightLimit))       // Increment.
       number++;
    else
        if ((direction <= -1) && (number > lowLimit))     // Decrement.
           number--;

    // Implement simple limit, and error check from conversion from text.
    if (number > heightLimit)
       number = heightLimit;
    else if (number < lowLimit)
            number = lowLimit;

    // Redisplay number into field
    sprintf (buffer, "%d", number);
    SetDlgItemText(hDlg, cntlID, buffer);

    // return number value.
    return number;
} 


// **********************************************************************************
// Function to process scroll bar messages and update spinner fields.
//
// hDlg   = handle to dialog box
// wParam = Scroll bar message
// lParam = control dialog handle
// 
// returns TRUE if spinner message was processed,
//         FALSE if it wasn't. 
BOOL ProcessSpinners (HWND hDlg, WORD wParam, LONG lParam)
{
    switch (wParam) // Find what type of action.
    {
         case SB_LINEDOWN:
            switch (GetDlgCtrlID(HIWORD(lParam))) /* handle of the control */
            {
                // Rotation settings.
                case IDC_ROTSPINX:
                    UpdateField (hDlg, IDC_ROTINPUTX, -1, RotationMax, RotationMin);
                    return TRUE;

                case IDC_ROTSPINY:
                    UpdateField (hDlg, IDC_ROTINPUTY, -1, RotationMax, RotationMin);
                    return TRUE;

                case IDC_ROTSPINZ:
                    UpdateField (hDlg, IDC_ROTINPUTZ, -1, RotationMax, RotationMin);
                    return TRUE;

                // Translation settings.
                case IDC_TRANSSPINX:
                    UpdateField (hDlg, IDC_TRANSINPUTX, -1, TranslationMax, TranslationMin);
                    return TRUE;

                case IDC_TRANSSPINY:
                    UpdateField (hDlg, IDC_TRANSINPUTY, -1, TranslationMax, TranslationMin);
                    return TRUE;

                case IDC_TRANSSPINZ:
                    UpdateField (hDlg, IDC_TRANSINPUTZ, -1, TranslationMax, TranslationMin);
                    return TRUE;

                // ZDistance.
                case IDC_ZDISTSPIN:
                    UpdateField (hDlg, IDC_ZDISTINPUT, -1, ZDistMax, ZDistMin);
                    return TRUE;

                // Object number
                case IDC_OBJNOSPIN:
                    UpdateField (hDlg, IDC_OBJNOINPUT, -1, ObjNoMax, ObjNoMin);
                    return TRUE;
            }
            break;

        case SB_LINEUP:
            switch (GetDlgCtrlID(HIWORD(lParam)))  /* handle of the control */
            {
                // Rotation settings.
                case IDC_ROTSPINX:
                    UpdateField (hDlg, IDC_ROTINPUTX, 1, RotationMax, RotationMin);
                    return TRUE;

                case IDC_ROTSPINY:
                    UpdateField (hDlg, IDC_ROTINPUTY, 1, RotationMax, RotationMin);
                    return TRUE;

                case IDC_ROTSPINZ:
                    UpdateField (hDlg, IDC_ROTINPUTZ, 1, RotationMax, RotationMin);
                    return TRUE;

                // Translation settings
                case IDC_TRANSSPINX:
                    UpdateField (hDlg, IDC_TRANSINPUTX, 1, TranslationMax, TranslationMin);
                    return TRUE;

                case IDC_TRANSSPINY:
                    UpdateField (hDlg, IDC_TRANSINPUTY, 1, TranslationMax, TranslationMin);
                    return TRUE;

                case IDC_TRANSSPINZ:
                    UpdateField (hDlg, IDC_TRANSINPUTZ, 1, TranslationMax, TranslationMin);
                    return TRUE;

                // ZDistance.
                case IDC_ZDISTSPIN:
                    UpdateField (hDlg, IDC_ZDISTINPUT, 1, ZDistMax, ZDistMin);
                    return TRUE;

                // Object number
                case IDC_OBJNOSPIN:
                    UpdateField (hDlg, IDC_OBJNOINPUT, 1, ObjNoMax, ObjNoMin);
                    return TRUE;
            }
            break;
    }

    return FALSE;
}

// **********************************************************************************
// Function used to retrieve the field values from the configuration dialog box,
// and place them into the INI configuration (global) variables. Used in
// save INI settings.
void GetFieldInfo (HWND hDlg)
{
    char buffer[StrSize];

    // Get 3D file name.
    GetDlgItemText(hDlg, IDC_3DFILENAME, buffer, StrSize);
    if (strlen (buffer) <= 0)
       strcpy (ObjFilename, "pyramid.txt");
    else
        strcpy (ObjFilename, buffer);

    // Rotation settings.
    XRot = UpdateField (hDlg, IDC_ROTINPUTX, 0, RotationMax, RotationMin);
    YRot = UpdateField (hDlg, IDC_ROTINPUTY, 0, RotationMax, RotationMin);
    ZRot = UpdateField (hDlg, IDC_ROTINPUTZ, 0, RotationMax, RotationMin);

    // Translation settings
    XTrans = UpdateField (hDlg, IDC_TRANSINPUTX, 0, TranslationMax, TranslationMin);
    YTrans = UpdateField (hDlg, IDC_TRANSINPUTY, 0, TranslationMax, TranslationMin);
    ZTrans = UpdateField (hDlg, IDC_TRANSINPUTZ, 0, TranslationMax, TranslationMin);

    // ZDistance.
    ZDistance = UpdateField (hDlg, IDC_ZDISTINPUT, 0, ZDistMax, ZDistMin);

     // Object number
     ObjectNum = UpdateField (hDlg, IDC_OBJNOINPUT, 0, ObjNoMax, ObjNoMin);
}


// ***************************************************************************
// *** Configuration Dialog Box.
BOOL FAR PASCAL _export ConfigDialog (HWND hDlg, WORD message, WORD wParam, LONG lParam)
{
    // used to read in configuration settings.
    char profileBuff[StrSize * 2];

    BOOL   success;
    int    bytesRead;

	switch (message)
    {
		case WM_INITDIALOG:
             // Read in INI settings and display these in fields.
             GetProfileString(pSzProgName, Section,
                              DefIniSettings, profileBuff, StrSize * 2);
             sscanf (profileBuff, "%d %d %d %d %d %d %d %d %s %s %d",
                                  &XRot, &YRot, &ZRot,
                                  &XTrans, &YTrans, &ZTrans,
                                  &ZDistance, &ObjectNum,
                                  ObjFilename, Password, &UsePassword);

             // Setup password settings.
             CheckDlgButton(hDlg, IDC_SETPASSWORD, UsePassword);
             EnableWindow(GetDlgItem(hDlg, IDC_LOADPWDIALOG), (BOOL) UsePassword);

             // Set Parameters into feilds.
             SetDlgItemInt(hDlg, IDC_ROTINPUTX, XRot, TRUE);
             SetDlgItemInt(hDlg, IDC_ROTINPUTY, YRot, TRUE);
             SetDlgItemInt(hDlg, IDC_ROTINPUTZ, ZRot, TRUE);

             SetDlgItemInt(hDlg, IDC_TRANSINPUTX, XTrans, TRUE);
             SetDlgItemInt(hDlg, IDC_TRANSINPUTY, YTrans, TRUE);
             SetDlgItemInt(hDlg, IDC_TRANSINPUTZ, ZTrans, TRUE);

             SetDlgItemInt(hDlg, IDC_ZDISTINPUT, ZDistance, TRUE);
             SetDlgItemInt(hDlg, IDC_OBJNOINPUT, ObjectNum, TRUE);

             SetDlgItemText(hDlg, IDC_3DFILENAME, ObjFilename);
			 return TRUE;

        // Process spinner (AKA: scroll bars) controls.
        case WM_VSCROLL:
            ProcessSpinners (hDlg, wParam, lParam);
            break;

        case WM_COMMAND:    // A user has selected a control command
            switch (LOWORD (wParam))
            {
                case IDC_SETPASSWORD:
                     if (UsePassword != 0)
                        UsePassword = 0;
                     else
                         UsePassword = 1;

                     CheckDlgButton(hDlg, IDC_SETPASSWORD, UsePassword);
                     EnableWindow(GetDlgItem(hDlg, IDC_LOADPWDIALOG), (BOOL) UsePassword);
                     break;

                case IDC_LOADPWDIALOG:
                     // Test Configuration Dialog.
                     static FARPROC fpfnSetPasswordDialog;
                     fpfnSetPasswordDialog = MakeProcInstance ((FARPROC)SetPasswordDialog, HInst); // Place in window constructor
                     DialogBox (HInst, MAKEINTRESOURCE(IDD_NEWPASSWORD), hDlg, (DLGPROC) fpfnSetPasswordDialog);
                     FreeProcInstance (fpfnSetPasswordDialog);
                     break;

                case IDC_SETIT:     // Save settings.
                     // Get user data.
                     GetFieldInfo (hDlg);

                     // Write settings to profile string.
                     // Rotations
                     sprintf (profileBuff, "%d %d %d ", XRot, YRot, ZRot);
                     // Translations
                     sprintf (profileBuff+strlen(profileBuff),
                              "%d %d %d ", XTrans, YTrans, ZTrans);
                     // zDist, ObjNo
                     sprintf (profileBuff+strlen(profileBuff),
                              "%d %d ", ZDistance, ObjectNum);
                     // 3d Filename, and password.
                     sprintf (profileBuff+strlen(profileBuff),
                              "%s %s %d ", ObjFilename, Password, UsePassword);

                     success = WriteProfileString(pSzProgName, Section,
                                                  profileBuff);
                     if (success == FALSE)
                     {
                        MessageBox(hDlg, "Unable to save settings!",
                                         "*** Write Profile String ***",
                                         MB_ICONSTOP | MB_OK | MB_TASKMODAL);
                        break;
                     }

				case IDC_FORGETIT:  // Don't save settings.
  					 EndDialog (hDlg, 0);
				     break;

                // Display file dialog box to retrieve 3D object file name.
                case IDC_3DFILEBROWSE:
                    char  szTemp  [StrSize]; 
                    char  szPath  [StrSize];                        // path+file name for OPENFILENAME.
                    char  szFile  [StrSize];                        // File name
                    char* szFilter[] = {"*.3D 3D Data Files", "*.3D", "*.* All Files", "*.*",""};
                    OPENFILENAME ofn;


                    GetDlgItemText(hDlg, IDC_3DFILENAME, szTemp, StrSize);
                    if (GetFileTitle(szTemp, szFile, StrSize) != 0)
                        szFile[0] = '\0';
                    strcpy (szPath, szTemp);
                    szPath[(strlen (szPath) - strlen(szFile))] = '\0';

                    memset(&ofn, 0, sizeof(OPENFILENAME) );     // reset structure.
                    ofn.lStructSize     = sizeof(OPENFILENAME);
                    ofn.hwndOwner       = hDlg;                 // owner is main window.
                    ofn.lpstrTitle      = "Steve's 3D File Browser Box";
                    ofn.lpstrFilter     = szFilter[0];
                    if (strlen (szFile) <= 0)                   // Test if there's a file name.
                        ofn.lpstrFile   = NULL;
                    else
                        ofn.lpstrFile   = szFile;               // file name buffer.
                    if (strlen (szPath) <= 0)                   // Test if there's a directory
                        ofn.lpstrInitialDir = NULL;
                    else
                        ofn.lpstrInitialDir = szPath;           // Start directory
                    ofn.nMaxFile        = StrSize;              // size of above.
                    ofn.lpstrFileTitle  = szFilter[1];          // file name buffer.
                    ofn.nMaxFileTitle   = 0;                    // size of above.
                    ofn.hInstance       = HInst;
                    ofn.Flags           = OFN_PATHMUSTEXIST | OFN_FILEMUSTEXIST |
                                          OFN_HIDEREADONLY;
    
                    if (GetOpenFileName(&ofn) != FALSE)
                    {
                        // place path in 3d file field.
                        SetDlgItemText(hDlg, IDC_3DFILENAME, szFile);
                    }
                    return TRUE;

				default:
                	break;
            }     
            break;

        case WM_CLOSE:
			EndDialog (hDlg, 0);
            break;
    }
              
	return FALSE;
}

void WaitMessy (HWND hWnd)
{
    MSG msg;

    while (PeekMessage(&msg, hWnd, 0, 0, PM_REMOVE));
}


// ***************************************************************************************
// ******************************* Screen Saver function *********************************
// **** Message Main dispatcher ! ****
long FAR PASCAL _export WndProc (HWND hWnd, WORD message, WORD wParam, LONG lParam)
{
    // Password checking routines
    const  LONG testPW = 0xffff;    // internal private message to init password entry. 
    static semiphore   = 0;         // Only one password entry can exist, this is incremented
                                    // to indicated such. (i.e. windows sends multiple useless
                                    // messages that required to method to be here).

    // Object variables.
    static Ball3D*       pMyBall = NULL;
    static ObjDistance*  pZDist  = NULL;
           char          profileBuff[StrSize * 2];
    static int           count;
    static RECT    oldRect, newRect, combRect;


    // Screen update variables.
    static HDC     virtualHDC;
    static HBITMAP hDrawBitmap, hOldBitmap;

    // Store last mouse positions, and check if it's changed.
    static POINT   lastMPos;

    switch (message)
    {                           
        // Window Constructor                
        case WM_CREATE:
             // Read in INI settings and display these in fields.
             GetProfileString(pSzProgName, Section,
                              DefIniSettings, profileBuff, StrSize * 2);
             sscanf (profileBuff, "%d %d %d %d %d %d %d %d %s %s %d",
                                  &XRot,   &YRot,   &ZRot,
                                  &XTrans, &YTrans, &ZTrans,
                                  &ZDistance, &ObjectNum,
                                  ObjFilename, Password, &UsePassword);

             // Allocate memory for objects .... only in local heap though.
             pMyBall = new Ball3D[ObjectNum];
             pZDist  = new ObjDistance[ObjectNum];

             if ((pMyBall == NULL) || (pZDist == NULL))
             {
                 delete pMyBall, pZDist; // One may not be allocated, but the other could.
                 MessageBox (hWnd, "Unable to Allocate Memory!", "*** Memory Error ***",
                             MB_OK | MB_TASKMODAL | MB_ICONSTOP);
                 PostQuitMessage (0);
                 return 0;
             }

             // intialise objects
             randomize();
             for (count = 0; count < ObjectNum; count++)
             {
                if (pMyBall[count].Load3DData (ObjFilename) != 0)
                {
                    delete pMyBall, pZDist;
                    MessageBox (hWnd, "Unable to Read 3D Data File!", ObjFilename,
                                MB_OK | MB_TASKMODAL | MB_ICONSTOP);
                    PostQuitMessage (0);
                    return 0;
                }

                pMyBall[count].Init (XTrans, YTrans, ZTrans,
                                    -(MAXX / 8), (MAXX / 8),
                                    -(MAXY / 4), (MAXY / 4),
                                     400,  ZDistance,
                                     random(XRot)+1, random(YRot)+1, random(ZRot)+1,
                                     MAXX >> 1,  MAXY >> 1);
             }

             // Create virtual screen and prepare it.
             HDC temp    = GetDC (hWnd);
             virtualHDC  = CreateCompatibleDC      (temp);
             SetMapMode (virtualHDC, MM_TEXT);
             hDrawBitmap = CreateCompatibleBitmap  (temp, MAXX, MAXY);
             hOldBitmap  = SelectObject (virtualHDC, hDrawBitmap);

             // Initialise dirty rectangle variable (structure).
             oldRect.left   = oldRect.top = 0;
             oldRect.right  = MAXX;
             oldRect.bottom = MAXY;
             
             FillRect(virtualHDC, &oldRect, GetStockObject(BLACK_BRUSH));
             ReleaseDC (hWnd, temp);

             // Get current mouse positions
             GetCursorPos(&lastMPos);
             ShowCursor (FALSE);
             break;

        case WM_TIMER:
             HDC realHDC = GetDC(hWnd); // Draw the object, get handle to real screen.

             // Move the objects into their new positions.
             for (count = 0; count < ObjectNum; count++)
             {
                pMyBall[count].Move();
                pZDist[count].objIndx      = count;
                pZDist[count].objZdistance = pMyBall[count].GetZDistance();
             }

             //**** Sort objects ****
             QuickSort (pZDist, 0L, ObjectNum-1);

             // Display objects by implementing simple dirty rectangle
             // screen updating techinque. 
             RectUD* pRectUpdate;
             newRect.left  = newRect.top = 0x7ffe;
             newRect.right = newRect.bottom = -1;
             for (count = ObjectNum-1; count >= 0; count--)
             {
                 pMyBall[pZDist[count].objIndx].Show(virtualHDC);

                 pRectUpdate = pMyBall[pZDist[count].objIndx].GetUpdateRegion();
                 // find max rectange to copy to real screen.

                 if (pRectUpdate -> xMin < newRect.left)
                    newRect.left = pRectUpdate -> xMin;

                 if (pRectUpdate -> xMax > newRect.right)
                    newRect.right = pRectUpdate -> xMax;

                 if (pRectUpdate -> yMin < newRect.top)
                    newRect.top = pRectUpdate -> yMin;

                 if (pRectUpdate -> yMax > newRect.bottom)
                    newRect.bottom = pRectUpdate -> yMax;
             }

             // Combine oldRect and newRect to get total area to copy
             // onto real screen. Implementation of simple dirty rectangle
             // method.
             if (oldRect.left < newRect.left)
                 combRect.left = oldRect.left;
             else
                 combRect.left = newRect.left;
                 
             if (oldRect.right > newRect.right)
                combRect.right = oldRect.right+1;
             else
                combRect.right = newRect.right+1;

             if (oldRect.top < newRect.top)
                combRect.top = oldRect.top;
             else
                 combRect.top = newRect.top;

             if (oldRect.bottom > newRect.bottom)
                combRect.bottom = oldRect.bottom+1;
             else
                combRect.bottom = newRect.bottom+1;

             // Blast it from virtual to real screen.
             BitBlt (realHDC,    combRect.left, combRect.top, (combRect.right - combRect.left), (combRect.bottom - combRect.top),
                     virtualHDC, combRect.left, combRect.top,
                     SRCCOPY);
             
             // Erase old positions in virtual screen
             // Adjust structure to erase right, and bottom borders inclusive.
             // Copy over new object rectangle.
             oldRect = newRect;
             newRect.right++;
             newRect.bottom++;
             FillRect(virtualHDC, &newRect, GetStockObject(BLACK_BRUSH));

             ReleaseDC   (hWnd, realHDC);
             UpdateWindow(hWnd);

             // Test to see if password dialog is open, and if so for how long?
             // If to long send password dialog box a close message.
             if (HPassDialog != NULL)
             {
                 CurrentCycle++;
                 if (CurrentCycle >= PASSWORDCYCLIMIT)
                 {
                     PostMessage (HPassDialog, WM_CLOSE, 0 ,0L);
                     CurrentCycle = 0; // Make sure don't send message twice if system is slow.
                 }
             }
             return TRUE;

        case WM_COMMAND:
             switch (lParam)
             {
                // Get user password...
                case testPW:
                     static FARPROC fpfnGetUserPDialog;
                     BOOL test;

                     ShowCursor (TRUE);
                     fpfnGetUserPDialog = MakeProcInstance ((FARPROC)GetUserPDialog, HInst);
                     test = DialogBox (HInst, MAKEINTRESOURCE(IDD_GETUSERPASSWORD), hWnd, (DLGPROC) fpfnGetUserPDialog);
                     FreeProcInstance (fpfnGetUserPDialog);
                     ShowCursor (FALSE);
                     if (test == 0)
                     {
                        // reset mouse pointer and return back to screen saver
                        GetCursorPos(&lastMPos);
                        SetActiveWindow (hWnd);
                     }
                     else
                         PostMessage (hWnd, WM_CLOSE, 0, 0L );
                     semiphore--;      // Free up semiphore.
                     return TRUE;
             }
             break;

        case WM_MOUSEMOVE:
             if ((LOWORD(lParam) == lastMPos.x) &&
                 (HIWORD(lParam) == lastMPos.y))
            {
                    break;
            }

        case WM_KEYDOWN:
        case WM_SYSKEYDOWN:
        case WM_LBUTTONDOWN:
        case WM_MBUTTONDOWN:
        case WM_RBUTTONDOWN:
             // Implement password check!
             if (semiphore == 0)
             {
                if ((UsePassword != 0) && (strcmp (Password, "~") != 0))
                {
                    semiphore++;    // Implement locking action.
                    PostMessage (hWnd, WM_COMMAND, 0, testPW);
                    break;
                }
                PostMessage (hWnd, WM_CLOSE, 0, 0L );
             }
             break;


        case WM_CLOSE:
        case WM_DESTROY:
             DeleteObject (SelectObject (virtualHDC, hOldBitmap));
             DeleteDC     (virtualHDC);
             delete pMyBall, pZDist;
             ShowCursor (TRUE);
             PostQuitMessage (0);
             return 0;
    }
    
    return DefWindowProc (hWnd, message, wParam, lParam);
}
