#ifndef _BALL3D_CODE
#define _BALL3D_CODE

#include <stdio.h>          // NULL, fopen(), fclose()
#include <stdlib.h>         // rand(), abs()
#include "3dobj.h"
#include "fix.h"            // NUMOFDEGREES
#include "ball3d.h"

// Debug
#include <conio.h>

// Constructor
Ball3D::Ball3D ()
{
}

void Ball3D::Init (int xSR, int ySR, int zSR,
                   int miX, int maX,
                   int miY, int maY,
                   int miZ, int maZ,
                   int xr,  int yr, int zr,
                   int xOrigin, int yOrigin)
{   // Assign 3D screen boundrys
    minX = miX; maxX = maX;
    minY = miY; maxY = maY;
    minZ = miZ; maxZ = maZ;

    // Assign rotation values.
    rotX = xr;
    rotY = yr;
    rotZ = zr;

    xSpdRng = xSR;
    ySpdRng = ySR;
    zSpdRng = zSR;

    // Initialise the x, y, z speed increment
    xSpd = (rand() % xSpdRng)+1;
    ySpd = (rand() % ySpdRng)+1;
    zSpd = (rand() % zSpdRng)+1;

    // set defaults for 3D object
    ballParams.sf = 1;
    ballParams.rx = 0;
    ballParams.ry = 0;
    ballParams.rz = 0;

    // Set the perspective parameters.
    ballParams.xOrigin = xOrigin;
    ballParams.yOrigin = yOrigin;
    ballParams.zDist   = minZ;

    // Adjust minZ to prevent division by 0
    minZ += (abs(maxZ - minZ) / 3);

    // These parameters will be used to move object over the screen
    ballParams.tx = 0;
    ballParams.ty = 0;
    ballParams.tz = minZ;              // set to the minimum distance
}

int Ball3D::Load3DData (char* pFileName)
{
    FILE* pFile = fopen (pFileName, "rt");

    if (pFile == NULL)
       return -1;

    object3D.Load3DObj (pFile);
    fclose (pFile);

    return 0;
}

void Ball3D::Show (HDC hdc)
{
    // Draw the object:
    object3D.Show (ballParams, hdc);
}

void Ball3D::Hide ()
{
    // Nothing here, don't know I should hide the object?
}

RectUD* Ball3D::GetUpdateRegion (void)
{
    return (object3D.GetUpdateRegion());
}

void Ball3D::Move (void)
{
    ballParams.rx += rotX;
    ballParams.ry += rotY;
    ballParams.rz += rotZ;

    // Check for 256 degree wrap around:
    if (ballParams.rx > NUMOFDEGREES-1)
       ballParams.rx = 0;

    if (ballParams.ry > NUMOFDEGREES-1)
       ballParams.ry = 0;

    if (ballParams.rz > NUMOFDEGREES-1)
           ballParams.rz = 0;

    //gotoxy (1,1);
    //printf ("x: %ld, y: %ld, z: %ld    ",  ballParams.tx, ballParams.ty, ballParams.tz);

    // Main working function.
    ballParams.tx += xSpd;
    ballParams.ty += ySpd;
    ballParams.tz += zSpd;

    // Process x coordinate
	if ((ballParams.tx <= minX) || (ballParams.tx >= maxX))
    {
       ballParams.tx += -xSpd;
       if (xSpd > 0)
       	   xSpd = -((rand() % xSpdRng)+1);
       else
           xSpd = ((rand() % xSpdRng)+1);
    }

    // Process y coordinate
    if ((ballParams.ty <= minY) || (ballParams.ty >= maxY))
    {
        ballParams.ty += -ySpd;
        if (ySpd > 0)
	       ySpd = -((rand() % ySpdRng)+1);
        else
            ySpd = ((rand() % ySpdRng)+1);

    }

    // Process z coordinate
    if ((ballParams.tz <= minZ) || (ballParams.tz >= maxZ))
    {
        ballParams.tz += -zSpd;
        if (zSpd > 0)
           zSpd = -((rand() % zSpdRng)+1);
        else
           zSpd = ((rand() % zSpdRng)+1);
    }
}

// Destrustor
Ball3D::~Ball3D ()
{
    // Nothing to destroy.
}

#endif // _BALL3D_CODE